/*
	Copyright (C) 2020-2022 Lewis ANESA.

	This file is part of ladspa-tool-kit

	ladspa-tool-kit is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	ladspa-tool-kit is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with ladspa-tool-kit.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <execinfo.h>
#include <sys/epoll.h>
#include "../include/instance.h"
#include "../include/utils.h"

/* General instance implementation */
static struct {
	LTKClassTypeDefinition_t *_class_type_definitions;

	LTKConfig_t _config;
} LTKroot = {
	0,
	{
		LTKCFG_ALLOC_ON_CLASS_CREATION,
		4,
		-1,
		-1,
		-1
	}
};

typedef struct LTKbuf_s {
	struct LTKbuf_s *_nxt;					/* Next pool */
	struct LTKClassTypeDefinition_s *_ctd;	/* Constructor */

	uint64_t _pcd;							/* Number of cumulated instance */
	uint64_t _fre[];						/* Bitewise array of free flags */
	/* Instances comes here after _fre */
} LTKbuf_t;

typedef struct LTKobj_s {
	LTKbuf_t *_buf;							/* Pool containing this instance */
	LTKFunction_t *_fct;					/* Function pointer to ctd's
												function array
											*/
	/* instance data */
} LTKobj_t;

struct LTKClassTypeDefinition_s {
	struct LTKClassTypeDefinition_s *_slv;	/* Slave */
	struct LTKClassTypeDefinition_s *_nxt;	/* Next */

	LTKbuf_t *_buf;							/* Pool list */
	uint64_t _cnt;							/* Highest instance id + 1 */
	uint64_t _dsz;							/* Data size multiple of
												sizeof(void *),
												including sizeof(LTKobj)
											*/
	LTKFreeHandler_t _inst_hdl;				/* Free instance handler */
	LTKFreeHandler_t _class_hdl;			/* Free class handler */
	void *_class_dat;						/* Free class data */
	LTKFunction_t _fct[];					/* Function array */
	/* Take care, function ID are absolute and positive,
		it is never relative to a derived instance
	*/
};

/* Give class of an object */
LTKClassTypeDefinition_t *LTKGetClass(void *instance) {
	LTKobj_t *obj = ((LTKobj_t *) instance) - 1;

	return obj->_buf->_ctd;
}

static uint64_t LTKCountMembers(LTKFunction_t * fct) {
	uint64_t ret;
	for(ret = 0; fct[ret]; ret++);

	return ret;
}

/* Run an instance method */
int LTKRun(void *instance, uint64_t fctid, ...) {
	va_list arg;
	va_start(arg, fctid);

	return LTKvRun(instance, fctid, arg);
}

/* Run an instance method */
int LTKvRun(void *instance, uint64_t fctid, va_list arg) {
	LTKobj_t *obj = ((LTKobj_t *) instance) - 1;

	if(!instance) {
		LTK_LOG_ERROR("Instance is missing");
		errno = EINVAL;

		return -1;
	}

	if(fctid >= LTKCountMembers(obj->_fct)) {
		LTK_LOG_ERROR("Function ID exceed instance's members");
		errno = EFAULT;

		return -1;
	}

	LTK_LOG_INFO("Running member %p::%p[no %" PRINT64 "]()",
		instance, obj->_fct[fctid], fctid
	);

	obj->_fct[fctid](obj + 1, arg);

	return 0;
}

/* Get buffer starting point */
static void * LTKGetBuffer(LTKbuf_t * buf) {
	return &buf->_fre[(buf->_fre[0] >> 7) +(buf->_fre[0] & 0x7E ? 2 : 1)];
}

/* Run a class method for each of it's instances */
static int LTKForEachClass(LTKClassTypeDefinition_t *ctd, int recursive,
		uint64_t fctid, LTKFunction_t callback, va_list arg) {
	LTKbuf_t *buf;
	enum {
		LTK_FOR_EACH_CLASS_NONE = 0,
		LTK_FOR_EACH_CLASS_FRIST_POOL = 1,
		LTK_FOR_EACH_CLASS_NEXT_POOL = 2
	} pool = LTK_FOR_EACH_CLASS_FRIST_POOL;

	LTK_LOG_INFO("Treating class %p", ctd);

	for(buf = ctd->_buf; buf;buf = (pool == LTK_FOR_EACH_CLASS_NEXT_POOL
			? (pool = LTK_FOR_EACH_CLASS_FRIST_POOL, ctd->_buf)
			: (pool = LTK_FOR_EACH_CLASS_NONE, buf->_nxt))) {
		uint64_t no;
		uint64_t cnt = buf->_fre[0];
		uint8_t *instance = LTKGetBuffer(buf);

		LTK_LOG_INFO("Treating pool %p", buf);

		for(no = 0; no < (cnt >> 7); no++) {
			uint8_t flag;

			if(!(~buf->_fre[no + 1])) {
				instance += ctd->_dsz << 6;
				continue;
			}

			for(flag = 0; flag < 64; flag++) {
				if(!(buf->_fre[no + 1] &(UINT64_C(1) << flag))) {
					va_list arg_copy;
					va_copy(arg_copy, arg);

					if(callback)
						callback((void *) &instance[sizeof(LTKobj_t)],
							arg_copy
						);
					else
						ctd->_fct[fctid]((void *) &instance[sizeof(LTKobj_t)],
							arg_copy
						);
				}

				instance += ctd->_dsz;

				/* check if member/callback has freed this instance
					and it's containing pool (only for first pool)
				*/
				if((pool == LTK_FOR_EACH_CLASS_FRIST_POOL)
						&& (ctd->_buf != buf)) {
					/* If yes, quickly switch to the next pool
						which is the ne first pool of this class
					*/
					pool = LTK_FOR_EACH_CLASS_NEXT_POOL;

					LTK_LOG_INFO("Pool %p freed during call", buf);

					break;
				}
			}

			if(pool == LTK_FOR_EACH_CLASS_NEXT_POOL) break;
		}

		if(pool == LTK_FOR_EACH_CLASS_NEXT_POOL) continue;

		for(no = 0; no <((cnt & 0x7E) >> 1); no++) {
			if(!(buf->_fre[(cnt >> 7) + 1] &(UINT64_C(1) << no))) {
				va_list arg_copy;
				va_copy(arg_copy, arg);

				if(callback)
					callback((void *) &instance[sizeof(LTKobj_t)],
						arg_copy
					);
				else
					ctd->_fct[fctid]((void *) &instance[sizeof(LTKobj_t)],
						arg_copy
					);
			}

			instance += ctd->_dsz;

			/* check if member/callback has freed this instance
				and it's containing pool (only for first pool)
			*/
			if((pool == LTK_FOR_EACH_CLASS_FRIST_POOL)
					&& (ctd->_buf != buf)) {
				/* If yes, quickly switch to the next pool
					which is the ne first pool of this class
				*/
				pool = LTK_FOR_EACH_CLASS_NEXT_POOL;

				break;
			}
		}
	}

	if(recursive) {
		int ret;
		LTKClassTypeDefinition_t *ptr;

		for(ptr = ctd->_slv; ptr; ptr = ctd->_nxt) {
			if(callback)
				ret = LTKForEachClass(ptr, 1, 0, callback, arg);
			else
				ret = LTKForEachClass(ptr, 1, fctid, 0, arg);

			if(ret < 0) return ret;
		}
	}

	return 0;
}

/* Run a class method for each of it's instances or it's subclass's instance */
int LTKRunForEach(LTKClassTypeDefinition_t *ctd,
		int recursive, uint64_t fctid, ...) {
	va_list arg;
	va_start(arg, fctid);

	return LTKvRunForEach(ctd, recursive, fctid, arg);
}

/* Run a class method for each of it's instances or it's subclass's instance */
int LTKvRunForEach(LTKClassTypeDefinition_t *ctd,
		int recursive, uint64_t fctid, va_list arg) {
	int ret;

	if(!ctd) {
		LTK_LOG_ERROR("Class is missing");
		errno = EINVAL;

		return -1;
	}

	if(fctid >= LTKCountMembers(ctd->_fct)) {
		LTK_LOG_ERROR("Function ID exceed class's members");
		errno = EFAULT;

		return -1;
	}

	LTK_LOG_INFO("Running member %p[no %" PRINT64
		"] on each instance of class %p%s",
		ctd->_fct[fctid], fctid, ctd,
		recursive ? " recursively" : ""
	);

	if((ret = LTKForEachClass(ctd, recursive, fctid, 0, arg)) < 0)
		return ret;

	return 0;
}

/* Run a class method for each of it's instances or it's subclass's instance */
int LTKForEach(LTKClassTypeDefinition_t *ctd, int recursive, LTKFunction_t callback, ...) {
	va_list arg;
	va_start(arg, callback);

	return LTKvForEach(ctd, recursive, callback, arg);
}

/* Run a class method for each of it's instances or it's subclass's instance */
int LTKvForEach(LTKClassTypeDefinition_t *ctd, int recursive, LTKFunction_t callback, va_list arg) {
	int ret;

	if(!ctd) {
		LTK_LOG_ERROR("Class is missing");
		errno = EINVAL;

		return -1;
	}

	if(!callback) {
		LTK_LOG_ERROR("Callback is missing");
		errno = EINVAL;

		return -1;
	}

	LTK_LOG_INFO("Running callback %p "
		"on each instance of class %p%s",
		callback, ctd,
		recursive ? " recursively" : ""
	);

	if((ret = LTKForEachClass(ctd, recursive, 0, callback, arg)) < 0)
		return ret;

	return 0;
}

LTKFunction_t LTKMember(LTKClassTypeDefinition_t *ctd, uint64_t fctid) {
	if(!ctd) {
		LTK_LOG_ERROR("Instance is missing");
		errno = EINVAL;

		return 0;
	}

	if(fctid >= LTKCountMembers(ctd->_fct)) {
		LTK_LOG_ERROR("Function ID exceed instance's members");
		errno = EFAULT;

		return 0;
	}

	return ctd->_fct[fctid];
}

/* Clear an instance content */
int LTKClear(void *instance) {
	LTKobj_t *obj = ((LTKobj_t *) instance) - 1;
	LTKbuf_t *buf;
	uint64_t dsz;

	if(!instance) {
		LTK_LOG_ERROR("Instance is missing");
		errno = EINVAL;

		return -1;
	}

	if(!(buf = obj->_buf)) {
		LTK_LOG_ERROR("Slot not yet instanciated or already freed");
		errno = EFAULT;

		return -1;
	}

	LTK_LOG_INFO("Clearing attributes of %p", instance);

	dsz = buf->_ctd->_dsz - sizeof(LTKobj_t);
	memset(instance, 0, dsz);

	return 0;
}

/* Allocate a buffer */
static LTKbuf_t *LTKBuf(LTKClassTypeDefinition_t * ctd, uint64_t cnt) {
	uint64_t no;
	LTKbuf_t *ret;

	if(!ctd) {
		LTK_LOG_ERROR("Class is missing");
		errno = EINVAL;

		return 0;
	}

	LTK_LOG_INFO("Adding %" PRINT64 " slot(s) to class %p", cnt, ctd);

	/* 0x10 is 16 octet space : 8 for fre[0] and 8 for fre[1] always allocated */
	ret = calloc(1, sizeof(LTKbuf_t) +
		0x10 +(((cnt - 1) >> 3) & ~(0x07 * UINT64_C(1))) +
		(ctd->_dsz * cnt)
	);

	if(!ret) {
		LTK_LOG_ERROR("Cannot allocte the pool");

		return 0;
	}

	ret->_fre[0] = (cnt << 1) | 1;
	ret->_ctd = ctd;
	if((ret->_nxt = ctd->_buf))
		ret->_pcd = ret->_nxt->_pcd;
	ret->_pcd += cnt;

	for(no = 0; no <(cnt >> 6); no++)
		ret->_fre[no + 1] = ~UINT64_C(0);

	for(no = 0; no <((ret->_fre[0] & 0x7E * UINT64_C(1)) >> 1); no++)
		ret->_fre[(cnt >> 6) + 1] |= UINT64_C(1) << no;

	return ret;
}

/* Create a constructor */
LTKClassTypeDefinition_t *LTKAddClass(LTKClassTypeDefinition_t *mtr,
		LTKFunction_t *fct, uint64_t dsz) {
	LTKClassTypeDefinition_t *ret;
	uint64_t fnb = 0;
	LTKFunction_t *fptr = 0;
	dsz += sizeof(LTKobj_t);

	/* Return error if not aligned instance */
	if(dsz % sizeof(int)) {
		LTK_LOG_ERROR("Data size(%" PRINT64 " octets) is not memory aligned",
			dsz
		);
		LTK_LOG_ERROR("Data size must be a multiple of %d", sizeof(int));
		errno = EINVAL;

		return 0;
	}

	/* Checking parent's functions and size */
	if(mtr) {
		fptr = mtr->_fct;

		while(fptr[0]) {
			fptr++;
			fnb++;
		}

		if(dsz < mtr->_dsz) {
			LTK_LOG_ERROR("New class can't contain less "
				"data than it's mother class"
			);
			LTK_LOG_ERROR("Mother class data size is %" PRINT64 " octets",
				mtr->_dsz
			);
			LTK_LOG_ERROR("New class data size is %" PRINT64 " octets", dsz);
			errno = EFAULT;

			return 0;
		}
	}
	else
		mtr = (LTKClassTypeDefinition_t *) &LTKroot;

	/* Function counting */
	if(fct) {
		fptr = &fct[fnb];

		while(fptr[0]) {
			fptr++;
			fnb++;
		}
	}

	/* Allocation */
	if(!(ret = (LTKClassTypeDefinition_t *)
			calloc(1, sizeof(LTKClassTypeDefinition_t)
			+ (sizeof(LTKFunction_t *) * (fnb + 1))))) {
		LTK_LOG_ERROR("Cannot allocte the class infos");

		return 0;
	}

	ret->_dsz = dsz;
	if(LTKroot._config._mask & LTKCFG_ALLOC_ON_CLASS_CREATION)
		ret->_buf = LTKBuf(ret, LTKroot._config._def_buf_size);
	ret->_nxt = mtr->_slv;
	mtr->_slv = ret;

	if(mtr != (LTKClassTypeDefinition_t *) &LTKroot) {
		ret->_inst_hdl = mtr->_inst_hdl;
		ret->_class_hdl = mtr->_class_hdl;
		ret->_class_dat = mtr->_class_dat;
	}

	/* Copying parent's methods if needed */
	{
		int x;
		LTKFunction_t *fmtr = 0;

		if(mtr) fmtr = mtr->_fct;
		fptr = fct;

		for(x = 0; x < fnb; x++) {
			if(fmtr && !fmtr[x]) fmtr = 0;
			if(fmtr && !fptr[x]) ret->_fct[x] = fmtr[x];
			else ret->_fct[x] = fptr[x];
		}

		ret->_fct[x] = 0;
	}

	return ret;
}
int LTKSetFreeInstanceCallback(LTKClassTypeDefinition_t *ctd
		, LTKFreeHandler_t hdl) {
	if(!ctd) {
		LTK_LOG_ERROR("Class is missing");
		errno = EINVAL;

		return -1;
	}

	ctd->_inst_hdl = hdl;

	return 0;
}

int LTKSetFreeClassCallback(LTKClassTypeDefinition_t *ctd
		, LTKFreeHandler_t hdl, void *dat) {
	if(!ctd) {
		LTK_LOG_ERROR("Class is missing");
		errno = EINVAL;

		return -1;
	}

	if(!hdl && dat) {
		LTK_LOG_ERROR("Cannot reset callback keeping data not null");
		errno = EINVAL;

		return -1;
	}

	ctd->_class_hdl = hdl;
	ctd->_class_dat = dat;

	return 0;
}

/* Free pool list */
static void LTKFreeBuf(LTKbuf_t *buf) {
	/* Recursive free */
	if(buf->_nxt) LTKFreeBuf(buf->_nxt);

	LTK_LOG_INFO("Removing pool %p of class %p", buf, buf->_ctd);

	/* If no handler, just free it directly */
	if(buf->_ctd->_inst_hdl) {
		uint64_t no;
		uint64_t cnt = buf->_fre[0];
		uint8_t *element = LTKGetBuffer(buf);

		for(no = 0; no < (cnt >> 7); no++) {
			int flag;

			if(!(~buf->_fre[no + 1])) {
				element += buf->_ctd->_dsz * 64;
				continue;
			}

			for(flag = 0; flag < 64; flag++) {
				if(!(buf->_fre[no + 1] & (UINT64_C(1) << flag))) {
					LTKobj_t *obj = (LTKobj_t *) element;
					void *ptr = (void *) &obj[1];

					obj->_buf = 0;
					buf->_ctd->_inst_hdl(ptr);
				}

				element += buf->_ctd->_dsz;
			}
		}

		for(no = 0; no <((cnt & 0x7E) >> 1); no++) {
			if(!(buf->_fre[(cnt >> 7) + 1] &(UINT64_C(1) << no))) {
				LTKobj_t *obj = (LTKobj_t *) element;
				void *ptr = (void *) &obj[1];

				obj->_buf = 0;
				buf->_ctd->_inst_hdl(ptr);
			}

			element += buf->_ctd->_dsz;
		}
	}

	free(buf);
}

/* Free class type definitions recursively */
static void LTKDtr(LTKClassTypeDefinition_t *ctd) {
	if(ctd->_slv) LTKDtr(ctd->_slv);
	if(ctd->_nxt) LTKDtr(ctd->_nxt);
	if(ctd->_buf) LTKFreeBuf(ctd->_buf);

	if(ctd->_class_hdl) ctd->_class_hdl(ctd->_class_dat);

	LTK_LOG_INFO("Removing class %p", ctd);

	free(ctd);
}

/* Delete instance */
int LTKFree(void *instance) {
	LTKobj_t *obj = ((LTKobj_t *) instance) - 1;
	LTKbuf_t *buf;
	LTKClassTypeDefinition_t *ctd;
	off_t array;
	off_t element;
	uint64_t no;
	uint64_t *fre;
	uint64_t flag;
	uint64_t num;

	/* Error check */
	if(!instance) {
		if(LTKroot._class_type_definitions) {
			LTK_LOG_INFO("Removing all LTK attached memory");
			LTKDtr(LTKroot._class_type_definitions);
		}
		else
			LTK_LOG_WARNING("Nothing to remove, no LTK attached memory");

		return 0;
	}

	if(!(buf = obj->_buf)) {
		LTK_LOG_ERROR("Instance already freed");
		errno = EFAULT;

		return -1;
	}

	/* Retrieve class type definition */
	ctd = buf->_ctd;
	/* Get array of slots */
	array = (off_t) LTKGetBuffer(buf);
	/* Point to current element */
	element = (off_t) obj;
	/* Retreive ID of pointed element */
	no = (element - array) / ctd->_dsz;
	/* Get, in array of free flags packed 64 by 64,
		the one containing this element
	*/
	fre = &buf->_fre[(no >> 6) + 1];
	/* Set to one the flag corresponding to this element */
	flag = UINT64_C(1) << (no & 0x3F);

	/* Checking memory corruption */
	if(fre[0] & flag) {
		LTK_LOG_WARNING("Instance already freed, "
			"correcting corrupted memory, please report"
		);

		obj->_buf = 0;
		errno = EFAULT;

		return -1;
	}

	/* Object has a handler to call before freeing memeory ?
		let's call it.
	*/
	if(ctd->_inst_hdl) ctd->_inst_hdl(instance);
	/* Make clear that element will be freed */
	obj->_buf = 0;

	/* Update free flag */
	fre[0] |= flag;
	/* Say that this pool has, at least, one free flag */
	buf->_fre[0] |= 1;
	num = no + (buf->_nxt ? buf->_nxt->_pcd : 0);

	/* If the "just freed" object isn't the "last one"
		e.g. the one with the biggest ID in this class.
		Just return, do nothing more.
	*/
	if((num + 1) != ctd->_cnt) return 0;

	/* Checking memory corruption */
	if(buf != ctd->_buf) {
		LTK_LOG_ERROR("Deeply corrupted memory, please report");

		errno = EFAULT;

		return -1;
	}

	LTK_LOG_INFO("Freed the lastest slot(no %" PRINT64 " of %p), "
		"searching the new one", num, ctd
	);

	/* If it is (the "last one") look for the new "last one" */
	flag >>= 1;
	num--;

	while(buf) {
		int free_found_buffer = 1;

		/*  */
		while(flag) {
			if(!(fre[0] & flag)) {
				ctd->_cnt = num + 1;
				return 0;
			}

			flag >>= 1;
			num--;
		}

		fre--;
		num--;

		while(fre > buf->_fre) {
			uint64_t free_bits = ~fre[0];

			if(free_bits) {
				int iter = 6;
				uint64_t dichoto[6] = {
					0xAAAAAAAAAAAAAAAA,
					0xCCCCCCCCCCCCCCCC,
					0xF0F0F0F0F0F0F0F0,
					0xFF00FF00FF00FF00,
					0xFFFF0000FFFF0000,
					0xFFFFFFFF00000000
				};

				while(iter--) {
					if(free_bits & dichoto[iter]) {
						free_bits &= dichoto[iter];
						continue;
					}

					free_bits &= ~dichoto[iter];
					num -= (1 << iter);
				}

				ctd->_cnt = num + 1;
				return 0;
			}

			fre--;
			num -= 64;
		}

		LTK_LOG_INFO("Pool %p totally freed, destroy it", buf);

		/* In case of last pool */
		if((ctd->_buf == buf) && !buf->_nxt) {
			LTKConfig_t config = {0};

			/* Proceed as config says */
			LTKConfigure(LTKCFG_GET | LTKCFG_FREE_LAST_POOL_IF_EMPTY, &config);
			free_found_buffer = config._mask ? 1 : 0;
		}

		/* Proceed if having to free actual buffer */
		if(free_found_buffer) {
			LTKbuf_t *nxt = buf->_nxt;
			buf->_nxt = 0; /* Cut LTKFreeBuf recursivity */
			ctd->_buf = nxt;
			LTKFreeBuf(buf);
			buf = nxt;
		}
		else return 0;

		/* Proceed if having to find a new free slot */
		if(buf) {
			/* Retreive ID of the very last element */
			no = (buf->_fre[0] >> 1) - 1;
			/* Get, in array of free flags packed 64 by 64,
				the one containing this element
			*/
			fre = &buf->_fre[(no >> 6) + 1];
			/* Set to one the flag corresponding to this element */
			flag = UINT64_C(1) <<(no & 0x3F);
		}
	}

	return 0;
}

/* Instanciate class defined by ctd */
void *LTKAdd(LTKClassTypeDefinition_t *ctd) {
	LTKobj_t *ret = 0;
	LTKbuf_t *buf;
	uint64_t num;

	if(!ctd) {
		LTK_LOG_ERROR("Class is missing");
		errno = EINVAL;

		return 0;
	}

	for(buf = ctd->_buf; buf; buf = buf->_nxt) {
		uint64_t no;
		uint64_t cnt = buf->_fre[0];
		uint8_t *element;
		uint8_t mflag;

		if(!(cnt & 1)) continue;

		LTK_LOG_INFO("Pool %p found with at least one free cell", buf);

		cnt >>= 1;
		element = LTKGetBuffer(buf);
		num = buf->_nxt ? buf->_nxt->_pcd : 0;

		for(no = 0; no <(cnt >> 6); no++) {
			if(buf->_fre[no + 1]) break;
			element += ctd->_dsz * 64;
		}

		if(no == (cnt >> 6)) {
			if(!buf->_fre[no + 1]) {
				buf->_fre[0] &= ~UINT64_C(1);
				continue;
			}

			mflag = cnt & 0x3FLLU;
		}
		else mflag = 64;

		cnt = no + 1;
		num += no << 6;

		for(no = 0; no < mflag; no++) {
			if(buf->_fre[cnt] &(UINT64_C(1) << no)) {
				ret = (LTKobj_t *) element;
				ret->_buf = buf;
				ret->_fct = ctd->_fct;

				buf->_fre[cnt] &= ~(UINT64_C(1) << no);

				num += no + 1;
				if(num > ctd->_cnt) ctd->_cnt = num;

				return ret + 1;
			}

			element += ctd->_dsz;
		}

		LTK_LOG_WARNING("Pool %p finally full corrupted memory correction, "
			"please report", buf
		);

		buf->_fre[0] &= ~UINT64_C(1);
	}

	/* No free pool, creating a new one, twice the size of the precedent if exist */
	num = ctd->_buf ? ctd->_buf->_pcd : 0;
	ctd->_cnt = num + 1;
	ctd->_buf =
	LTKBuf(ctd, ctd->_buf
		? ctd->_buf->_fre[0]
		: LTKroot._config._def_buf_size
	);
	ctd->_buf->_fre[1] &= ~UINT64_C(1);

	ret = (LTKobj_t *) LTKGetBuffer(ctd->_buf);
	ret->_buf = ctd->_buf;
	ret->_fct = ctd->_fct;

	return ret + 1;
}

/* Get instance number in pools */
uint64_t *LTKGetObjectID(void *instance) {
	LTKobj_t *obj = ((LTKobj_t *) instance) - 1;
	LTKbuf_t *buf;
	LTKClassTypeDefinition_t *ctd;
	off_t array;
	off_t element;
	static uint64_t no;

	if(!instance) {
		LTK_LOG_ERROR("Instance is missing");
		errno = EINVAL;

		return 0;
	}

	if(!(buf = obj->_buf)) {
		LTK_LOG_ERROR("Slot not yet instanciated or already freed");
		errno = EFAULT;

		return 0;
	}

	ctd = buf->_ctd;
	array = (off_t) LTKGetBuffer(buf);
	element = (off_t) obj;
	no = (element - array) / ctd->_dsz;
	if(buf->_nxt) no += buf->_nxt->_pcd;
	return &no;
}

/* Get instance from it's number in the pool */
void *LTKGetObjectFromID(LTKClassTypeDefinition_t * ctd, uint64_t id) {
	LTKbuf_t *buf;

	if(!ctd) {
		LTK_LOG_ERROR("Class is missing");
		errno = EINVAL;
		return 0;
	}

	if(!(buf = ctd->_buf)) {
		LTK_LOG_ERROR("No slot created at this point");
		errno = EFAULT;

		return 0;
	}

	if(id > buf->_pcd) {
		LTK_LOG_ERROR("Slot ID %" PRINT64
		" exceeds number of slots for this class", id);
		errno = EFAULT;

		return 0;
	}

	while(buf) {
		LTK_LOG_INFO("Searching id %" PRINT64 "in pool %p", id, buf);

		if(!buf->_nxt) {
			uint8_t *element;

			if(buf->_fre[(id >> 6) + 1] &(UINT64_C(1) <<(id & 0x3FLLU))) {
				LTK_LOG_ERROR("Slot ID %" PRINT64
					" not yet instanciated or already freed", id
				);
				errno = ENOMEM;

				return 0;
			}

			element = LTKGetBuffer(buf);
			return((LTKobj_t *) & element[id * ctd->_dsz]) + 1;
		}

		if(id >= buf->_nxt->_pcd) {
			if(id < buf->_pcd) {
				uint8_t *element;
				id -= buf->_nxt->_pcd;

				if(buf->_fre[(id >> 6) + 1] &(UINT64_C(1) <<(id & 0x3F))) {
					LTK_LOG_ERROR("Slot ID %" PRINT64
						" not yet instanciated or already freed", id
					);
					errno = ENOMEM;

					return 0;
				}

				element = LTKGetBuffer(buf);
				return((LTKobj_t *) & element[id * ctd->_dsz]) + 1;
			}
		}

		buf = buf->_nxt;
	}

	LTK_LOG_ERROR("Internal error, please report");
	errno = EADV;

	return 0;
}

/* Control function */
int LTKConfigure(uint64_t flags, LTKConfig_t *config) {
	int x;
	int ret = 0;

	if(!config) {
		LTK_LOG_ERROR("Config is missing");
		errno = EINVAL;

		return -1;
	}

	if(flags & LTKCFG_SET) {
		LTKroot._config._mask =
		(LTKroot._config._mask & ~flags) | (config->_mask & flags);
	}
	else
		config->_mask =
		((LTKroot._config._mask & flags) | (config->_mask & ~flags))
		& LTKCFG_GET_ALL;

	for(x = 1; x < 64; x++) {
		uint64_t flag = UINT64_C(1) << x;

		if(!(flags & flag))
		continue;

		switch(flag) {
		case LTKCFG_SIZE_ON_CLASS_CREATION:
			if(flags & LTKCFG_SET) {
				if(config->_def_buf_size)
					LTKroot._config._def_buf_size = config->_def_buf_size;
				else {
					errno = EFAULT;
					ret = 1;
				}
			}
			else
				config->_def_buf_size = LTKroot._config._def_buf_size;
			continue;

		case LTKCFG_LOG_FD_ERROR:
			if(flags & LTKCFG_SET)
				LTKroot._config._log_fd_error = config->_log_fd_error;
			else
				config->_log_fd_error = LTKroot._config._log_fd_error;
			continue;

		case LTKCFG_LOG_FD_WARNING:
			if(flags & LTKCFG_SET)
				LTKroot._config._log_fd_warning = config->_log_fd_warning;
			else
				config->_log_fd_warning = LTKroot._config._log_fd_warning;
			continue;

		case LTKCFG_LOG_FD_INFO:
			if(flags & LTKCFG_SET)
				LTKroot._config._log_fd_info = config->_log_fd_info;
			else
				config->_log_fd_info = LTKroot._config._log_fd_info;
			continue;

		default:;
		}
	}

	return ret;
}

int LTKLog(unsigned int log_level, char *file, int line, const char *func,
		char *format, ...) {
	size_t ret;
	va_list arg;

	struct {
		int _fd;
		char *_color;
		char *_type;
	} conf[3] = {
		{LTKroot._config._log_fd_error, "\033[31m", "err"},
		{LTKroot._config._log_fd_warning, "\033[33m", "wrn"},
		{LTKroot._config._log_fd_info, "\033[32m", "inf"}
	};

	if((log_level > 2) || (conf[log_level]._fd < 0)) return 0;

	/* Unfortunately, can't log from log function itself */
	if((ret = dprintf(conf[log_level]._fd, "%sLTK log <%s> at %s:%d in ",
			conf[log_level]._color, conf[log_level]._type, file, line)) < 0)
		return -1;

	ret += dprintf(conf[log_level]._fd, "%s : ",
		func ? func : "not mentionned function"
	);
	va_start(arg, format);
	ret += vdprintf(conf[log_level]._fd, format, arg);
	va_end(arg);
	ret += dprintf(conf[log_level]._fd, "\033[0m\n");

	return ret;
}

